//==========================================================================
// File: demo.c
// Description: Test subdivision traversal algorithms
// Author: Pat Morin
// Created: Mon Mar 29 19:09:25 EST 1999
// Notes: #define MEASURING if you want to measure comparisons instead of
//        running time.
//==========================================================================
#include <iostream.h>
#include <ctype.h>
#include <math.h>
#include <time.h>

#include <LEDA/list.h>
#include <LEDA/point.h>
#include <LEDA/circle.h>
#include <LEDA/random_source.h>
#include <LEDA/graph.h>
#include <LEDA/graph_alg.h>
#include <LEDA/graph_misc.h>
#include <LEDA/graph_gen.h>
#include <LEDA/plane_alg.h>

// #define WIN

#ifdef WIN
#include <LEDA/window.h>
#endif // WIN

#include "util.h"
#include "mesh.h"

#include "traversal.h"

/*
 * Various types of test graphs
 */
enum graphtype { G_DELAUNAY, G_GRAHAM, G_MESH, G_MULMULEY };


/*
 * These are colors we use to label connected components in our display
 */
#ifdef WIN
#define NCOLS 16
color colors[NCOLS] =
{ black, white, red, green, blue, yellow, violet, orange, cyan, 
  brown, pink, green2, blue2, grey1, grey2, grey3 };
#endif // WIN

/* 
 * Run the tests and ouput the results in tabular form.  The results
 * for each graph are output on a single line in the form:
 * <bkootime> <bkoocmp> <bmtime> <bmcmp> <dfstime>
 */
int main(int argc, char *argv[]) 
{
  node v;
  edge e;

  // read command line arguments
  if (argc != 6) {
    cerr << "Usage: " << argv[0] << " <graph-type> <n> <vfaults>"
	 << " <efaults> <nreps>" << endl
	 << " Where <graph-type> is one of delaunay, graham, mesh, or segment." 
	 << endl;
    exit(1);
  }
  int n = atoi(argv[2]);
  graphtype gtype;
  switch (toupper(argv[1][0])) {
  case 'D':
    gtype = G_DELAUNAY;
    break;
  case 'G':
    gtype = G_GRAHAM;
    break;
  case 'M':
    n = (int) ceil(sqrt(n));
    n *= n;
    gtype = G_MESH;
    break;
  case 'S':
    gtype = G_MULMULEY;
    break;
  default:
    cerr << "Invalid graph type" << endl;
    exit(1);
    break;
  }
  int vfaults = atoi(argv[3]);
  int efaults = atoi(argv[4]);
  int nreps = atoi(argv[5]);

  // create our persistent data structures
  random_source R;
  GRAPH<point,int> G;
  list<point> L;

#ifdef WIN
    // open the display window
    window W(300, 300, "unit graph debug window");  
    W.set_bg_color(ivory);
    W.open(window::center,window::center);
    W.init(-.1, 1.1, -.1);
#endif // WIN

  // we do nreps repetitions of the same experiment
  for (int round = 0; round < nreps; round++) {

    // generate a random test graph
    L.clear();
    double x, y;
    for (int i = 0; i < n; i++) {
      do {
	R >> x;
	R >> y;
      } while (point(0.5, 0.5).distance(point(x,y)) > 0.5);
      L.append(point(x, y));
    }
    switch(gtype) {
    case G_DELAUNAY:
      DELAUNAY_TRIANG(L, G);
      break;
    case G_GRAHAM:
      TRIANGULATE_POINTS(L, G);
      break;
    case G_MESH:
      MESH(n, G);
      break;
    case G_MULMULEY:
      // make our points into the line segments
      list<segment> L2;
      while (!L.empty()) {
	point p1 = L.head();
	L.del(L.first());
	if (!L.empty()) {
	  point p2 = L.head();
	  L.del(L.first());
	  L2.append(segment(p1, p2));
	}
      }
      // compute hte arrangement of line segments
      GRAPH<point,segment> GS; 
      MULMULEY_SEGMENTS(L2, GS, true);

      // do a type conversion for our functions
      node_array<node> vmap(GS);
      G.clear();
      forall_nodes (v, GS) {  // first the nodes
	node vp = G.new_node();
	G[vp] = GS[v];
	vmap[v] = vp;
      }
      forall_nodes (v, GS) { // then the edges
	edge e;
	forall_adj_edges (e, v) {
	  G.new_edge(vmap[GS.source(e)], vmap[GS.target(e)]);
	}
      }
      GS.clear();
      G.make_planar_map();
      break;
    }

    // add vertex faults
    list<node> V;
    V = G.all_nodes();
    V.permute();
    for (int i = 0; i < vfaults; i++) {
      v = V[V.first()];
      V.del(V.first());
      G.del_node(v);
    }
    // add edge faults
    list<edge> E;
    E = G.all_edges();
    // E.permute();
    for (int i = 0; i < efaults; i++) {
      e = E[E.first()];
      E.del(E.first());
      G.del_edge(G.reversal(e));
      G.del_edge(e);
    }
    G.compute_faces();
    node_array<int> compnum(G);
    COMPONENTS(G, compnum);

#ifdef WIN
    // display the graph with color-coded components
    W.clear();
    forall_edges (e, G) {
      W.draw_segment(G[G.source(e)], G[G.target(e)], 
		     colors[compnum[G.source(e)]%NCOLS]);
    }
#endif // WIN

    edge estart;
    int compsize, ntries = 0;
    do {
      compsize = 0;
      estart = G.choose_edge();
      forall_nodes (v, G) {
	if (compnum[v] == compnum[G.source(estart)])
	  compsize++;
      }
      ntries++;
    } while (ntries <= 10 && compsize < G.number_of_nodes() * 98 / 100);
    if (ntries > 10) continue;

#ifdef MEASURING
    bkoo_cmp_count = 0;
#endif // MEASURING
    clock_t sbkoo = clock();
    bkoo_traversal(G, estart);
    clock_t fbkoo = clock();

#ifdef MEASURING
    bm_cmp_count = 0;
#endif // MEASURING
    clock_t sbm = clock();
    bm_traversal(G, estart);
    clock_t fbm = clock();

    clock_t sdfs = clock();
    dfs_traversal(G, estart);
    clock_t fdfs = clock();
    
#ifdef MEASURING
    G.clear();

    cout << (double)(fbkoo-sbkoo)/CLOCKS_PER_SEC 
	 << " " << bkoo_cmp_count << " "
	 << (double)(fbm-sbm)/CLOCKS_PER_SEC << " " << bm_cmp_count << " "
	 << (double)(fdfs-sdfs)/CLOCKS_PER_SEC << endl;
#else
    cout << (double)(fbkoo-sbkoo)/CLOCKS_PER_SEC << " " 
	 << (double)(fbm-sbm)/CLOCKS_PER_SEC << " " 
	 << (double)(fdfs-sdfs)/CLOCKS_PER_SEC << endl;
#endif // MEASURING
  }
  return 0;
}

