//==========================================================================
// File: demo.c
// Description: Test subdivision traversal algorithms on GIS type files
// Author: Pat Morin
// Created: Mon Mar 29 19:09:25 EST 1999
//==========================================================================
#include <iostream.h>
#include <fstream.h>
#include <ctype.h>
#include <math.h>
#include <time.h>

#include <LEDA/list.h>
#include <LEDA/point.h>
#include <LEDA/circle.h>
#include <LEDA/rat_point.h>
#include <LEDA/rat_segment.h>
#include <LEDA/random_source.h>
#include <LEDA/graph.h>
#include <LEDA/graph_alg.h>
#include <LEDA/graph_misc.h>
#include <LEDA/graph_gen.h>
#include <LEDA/plane_alg.h>

#define WIN

#ifdef WIN
#include <LEDA/window.h>
#endif // WIN

#include "util.h"
#include "report.h"

#include "traversal.h"

/*
 * These are colors we use to label connected components in our display
 */
#ifdef WIN
#define NCOLS 16
color colors[NCOLS] =
{ black, white, red, green, blue, yellow, violet, orange, cyan, 
  brown, pink, green2, blue2, grey1, grey2, grey3 };
#endif // WIN


GRAPH<point,int> *pG;

int dir_cmp(const edge &e1, const edge &e2)
{
  double d1 = ray(pG->inf(pG->source(e1)), 
		  pG->inf(pG->target(e1))).direction();
  double d2 = ray(pG->inf(pG->source(e2)), 
		  pG->inf(pG->target(e2))).direction();
  return compare(d1, d2);
}

int compare(const rat_segment &s1, const rat_segment &s2)
{
  if (s1.source().xcoord() < s2.source().xcoord())
    return -1;
  if (s1.source().xcoord() > s2.source().xcoord())
    return 1;
  if (s1.source().ycoord() < s2.source().ycoord())
    return -1;
  if (s1.source().ycoord() > s2.source().ycoord())
    return 1;
  if (s1.target().xcoord() < s2.target().xcoord())
    return -1;
  if (s1.target().xcoord() > s2.target().xcoord())
    return 1;
  if (s1.target().ycoord() < s2.target().ycoord())
    return -1;
  if (s1.target().ycoord() > s2.target().ycoord())
    return 1;
  return 0;
}


/* Read a shapefile-like thing and make it into a map */
void READ_SHAPE_FILE(GRAPH<point,int> &G, istream &is)
{
  list<rat_segment> LS;
  int npolys;

  // convert everything to segments
  is >> npolys;
  for (int i = 0; i < npolys; i++) {
    list<point> L;
    L.read(is, '\n');
    list_item it;
    if (L.size() > 20) {
      forall_items (it, L) {
	rational x1(L[it].xcoord());
	rational x2(L[L.cyclic_succ(it)].xcoord());
	rational y1(L[it].ycoord());
	rational y2(L[L.cyclic_succ(it)].ycoord());
	LS.append(rat_segment(x1, y1, x2, y2));
      }
    }
    cout << LS.size() << endl;
    if (LS.size() > 100000) break;
  }
  
  // get rid of duplicate segments
  LS.sort();
  LS.unique();

  // Create a planar map
  GRAPH<rat_point,rat_segment> GR;
  MULMULEY_SEGMENTS(LS, GR, true);

  // copy GR into G
  G.clear();
  node_array<node> vmap(GR);
  G.clear();
  node v;
  forall_nodes (v, GR) {  // first the nodes
    node vp = G.new_node();
    G[vp] = GR[v].to_point();
    vmap[v] = vp;
  }
  forall_nodes (v, GR) { // then the edges
    edge e;
    forall_adj_edges (e, v) {
      G.new_edge(vmap[GR.source(e)], vmap[GR.target(e)]);
    }
  }
  GR.clear();
  pG = &G;
  G.sort_edges(dir_cmp);
  G.make_planar_map();
  G.compute_faces();
}

/* 
 * Run the tests and ouput the results in tabular form.  The results
 * for each graph are output on a single line in the form:
 * <bkootime> <bkoocmp> <bmtime> <bmcmp> <dfstime>
 */
int main(int argc, char *argv[]) 
{
  node v;
  edge e;

  // read command line arguments
  if (argc != 2) {
    cerr << "Usage: " << argv[0] << " <shapefile>" << endl;
    exit(1);
  }

  // open input file and make it into a graph
  ifstream is(argv[1]);
  if (is.fail()) {
    cerr << "Unable to open input file " << argv[1] << endl;
    exit(1);
  }
  GRAPH<point,int> G;
  READ_SHAPE_FILE(G, is);
  G.make_planar_map();
  G.compute_faces();
  node_array<int> compnum(G);
  int nc = COMPONENTS(G, compnum);
  cout << "G has " << G.number_of_nodes() << " vertices, "
       << G.number_of_edges()/2 << " edges, "
       << G.number_of_faces() << " faces, and " 
       << nc << " component(s)." << endl;
#ifdef WIN
  // display the graph with color-coded components
  window W(600, 600, "unit graph debug window");  
  W.set_bg_color(ivory);
  W.open(window::center,window::center);
  double xmin = HUGE_VAL, xmax = -HUGE_VAL, 
    ymin = HUGE_VAL, ymax = -HUGE_VAL;
  forall_nodes (v, G) {
    if (G[v].xcoord() < xmin) xmin = G[v].xcoord();
    if (G[v].xcoord() > xmax) xmax = G[v].xcoord();
    if (G[v].ycoord() < ymin) ymin = G[v].ycoord();
    if (G[v].ycoord() > ymax) ymax = G[v].ycoord();
  }
  W.init(xmin, xmax, ymin);
  W.clear();
  forall_edges (e, G) {
    W.draw_segment(G[G.source(e)], G[G.target(e)], 
		   colors[compnum[G.source(e)]%NCOLS]);
  }
  W.set_node_width(1);
  forall_nodes (v, G) {
    W.draw_filled_node(G[v], red);
  }


#endif // WIN
  // we do nreps repetitions of the same experiment
  int nreps = 10;
  for (int round = 0; round < nreps; round++) {
    edge estart = G.choose_edge();

    nv = 0; ne = 0; nf = 0;
    clock_t sbkoo = clock();
    bkoo_traversal(G, estart);
    clock_t fbkoo = clock();
    cout << "bkoo reported " << nv << " " << ne << " " << nf << endl;

    nv = 0; ne = 0; nf = 0;
    clock_t sbm = clock();
    bm_traversal(G, estart);
    clock_t fbm = clock();
    cout << "bm reported " << nv << " " << ne << " " << nf << endl;

    nv = 0; ne = 0; nf = 0;
    clock_t sdfs = clock();
    dfs_traversal(G, estart);
    clock_t fdfs = clock();
    cout << "dfs reported " << nv << " " << ne << " " << nf << endl;
    
    cout << (double)(fbkoo-sbkoo)/CLOCKS_PER_SEC << " " 
	 << (double)(fbm-sbm)/CLOCKS_PER_SEC << " " 
	 << (double)(fdfs-sdfs)/CLOCKS_PER_SEC << endl;
  }
  return 0;
}


